package zb2smt;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;
import net.sourceforge.czt.z.ast.Pred;
import network.client.ProofClient;
import zb2smt.bol.Predicate;
import zb2smt.convertZ2B.ConvertCircus2Bol;

/**
 * 
 * @author Alessandro Gurgel, Valerio Junior
 *
 */
public class ZB2SMTUtils {

	public static final String smtSolversStr = "smt-solvers";
	public static final String fs = File.separator;

	private static final String veriTproverName = "veriT";
	private static final String veriTpathWin = fs + smtSolversStr + fs + veriTproverName + fs + "win" + fs + "rv.exe " ;
	private static final String veriTpathLinux = fs  +  smtSolversStr + fs + veriTproverName + fs + "linux" + fs + "rv ";

	private static final String z3proverName = "z3";
	private static final String z3path = fs  +  smtSolversStr + fs + z3proverName + fs + "z3.exe ";

	private static final String cvc3proverName = "cvc3";
	private static final String cvc3path = fs  +  smtSolversStr + fs + cvc3proverName + fs + "cvc3-optimized -lang smtlib ";

	private static final String altErgoproverName = "alt-ergo";
	private static final String altErgopath = fs  +  smtSolversStr + fs + altErgoproverName + fs + "alt-ergo0.9 ";

	private static final String yicesProverName = "yices";
	private static final String yicesPathWin = fs + smtSolversStr + fs + yicesProverName + fs + "win" + fs + "yices.exe -smt ";
	private static final String yicesPathLinux = fs + smtSolversStr + fs + yicesProverName + fs + "linux" + fs + "yices.exe -smt ";

	private static final String generatedFilePath = fs + smtSolversStr + fs; //Gera os arquivos na raiz do projeto

	private static final String abspath = (new File("")).getAbsolutePath();

	private static final String veriTparamPureSMTFileCmd = " --print-simp-and-exit ";

	private static final String lineSep = System.getProperty("line.separator");

	private static final String WIN_ID = "Windows";
	private static final String defaultStr = "\\bin\\zb2smt";


	/**
	 * Esse método pega um predicado em Circus e gera o correspondente
	 * no path passado como parametro. 
	 * @return true se o predicado pode ser passado para Predicado
	 * @param pred Predicado Z ou Circus
	 * @param pathFile arquivo onde deve ser gerado.
	 */
	public static boolean generateVeritFile(String pathFile, Pred pred){
		/*
		 * Convertendo Circus 2 Bol 
		 */
		ConvertCircus2Bol convert = new ConvertCircus2Bol();

		Predicate pred2 = (Predicate)convert.visitTerm(pred);
		if (pred2 != null) {
			POLabel po = new POLabel("PO", pred2, null, convert.getIdentificadores(), null);
			LinkedList<POLabel> pos = new LinkedList<POLabel>();
			pos.add(po);

			generateVeritFile(pathFile, pos);
			return true;
		}
		else{
			return false;
		}
	}
	//TODO : Verificar se esse metodo serve para vc, Valerio. Lembrando que eu tb
	//utilizo ele
	/**
	 * Utilizado diretamente por B
	 * @param pathFile arquivo onde deve ser gerado.
	 * @param pos lista de POLabels utilizadas no printer do Harvey
	 */
	public static void generateVeritFile(String pathFile, LinkedList<POLabel> pos) {

		POPrintVerit harveyPrinter = new POPrintVerit();

		StringBuffer body = new StringBuffer();
		String indent = "    ";

		harveyPrinter.toPrintInBuffer(pos, body, indent, "file");

		/* Criacao do Arquivo*/

		File output;
		if (pathFile.endsWith(".smt")) {
			output = new File(pathFile);
		}
		else{
			output = new File(pathFile + ".smt");
		}
		try {
			FileOutputStream out = new FileOutputStream(output);
			byte [] data = body.toString().getBytes();
			out.write(data);
			out.close();

		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}


	}

	private static ResultStatusOP executeMultiPlataformProver(String pathFile, String pathLinux, String pathWin){

		ResultStatusOP result = ResultStatusOP.UNDETERMINED;
		/* Verifica se é Windows*/
		String os = System.getProperty("os.name");
		String cmd;


		if ( os!= null && os.startsWith(WIN_ID)){
			pathFile = "\"" + pathFile + "\"";
			cmd = /*getAbsPathforWindows()*/ abspath + pathWin + pathFile;
		}
		else {
			/*
			 * Para que o harvey tenha permissao de ser executado
			 */

			try {
				executeCmd("chmod +x " + abspath + pathLinux);
			} catch (IOException e) {
				e.printStackTrace();
			}

			cmd = abspath + veriTpathLinux + pathFile;
		}

		try {

			//System.out.println((new File("")).getAbsolutePath());
			String output = executeCmd(cmd);
			String[] lines = returnLines(output);

			String respostaHarvey = "";
			if (lines.length > 0) {
				respostaHarvey = lines[lines.length - 1];
			}
			return getResultFromSolversAnswerString(respostaHarvey);

		} catch (IOException e) {
			e.printStackTrace();
		}

		return result;
	}


	private static ResultStatusOP executeZ3(String pathFile){

		ResultStatusOP result = ResultStatusOP.UNDETERMINED;
		/* Verifica se é Windows*/
		String cmd;

		pathFile = "\"" + pathFile + "\"";
		cmd = /*getAbsPathforWindows()*/ abspath + z3path + pathFile;
		try {
			String output = executeCmd(cmd);
			String[] lines = returnLines(output);

			String respostaHarvey = "";
			if (lines.length > 0) {
				respostaHarvey = lines[lines.length - 1];
			}
			return getResultFromSolversAnswerString(respostaHarvey);

		} catch (IOException e) {
			e.printStackTrace();
		}

		return result;
	}


	private static ResultStatusOP executeLinuxProver(String pathFile, String smtproverPath){

		ResultStatusOP result = ResultStatusOP.UNDETERMINED;
		/* Verifica se é Windows*/
		String cmd;
		/*
		 * Para que o harvey tenha permissao de ser executado
		 */

		try {
			executeCmd("chmod +x " + abspath + smtproverPath);
		} catch (IOException e) {
			e.printStackTrace();
		}

		cmd = abspath + veriTpathLinux + pathFile;

		try {

			String output = executeCmd(cmd);
			String[] lines = returnLines(output);

			String respostaHarvey = "";
			if (lines.length > 0) {
				respostaHarvey = lines[lines.length - 1];
			}
			return getResultFromSolversAnswerString(respostaHarvey);

		} catch (IOException e) {
			e.printStackTrace();
		}

		return result;
	}



	/**
	 * 
	 * @param str string answer from a smt solver
	 * @return a resultStatusOP
	 */
	private static ResultStatusOP getResultFromSolversAnswerString (String str) {
		if (str.equals(SMTString.UNSAT)) {
			return ResultStatusOP.CHECKED_TRUE;
		} else if (str.equals(SMTString.SAT)) {
			return ResultStatusOP.CHECKED_FALSE;
		} else if (str.equals(SMTString.UNKNOWN)) {
			return ResultStatusOP.UNDETERMINED;
		}
		return ResultStatusOP.SINTATIC_ERROR;
	}

	/**
	 * Metodo auxiliar de chamada de comandos em terminais
	 * @param cmd
	 * @return
	 * @throws IOException
	 */
	private static String executeCmd(String cmd)
	throws IOException
	{
		Process p = Runtime.getRuntime().exec(cmd);
		// Get the input stream and read from it

		InputStream in = p.getInputStream();
		InputStream error = p.getErrorStream();

		String output = "";
		String erro = "";
		int c;
		if (error != null) {
			while ((c = in.read()) != -1) {
				output = output + ((char)c);
			}
			in.close();

			while ((c = error.read()) != -1) {
				erro = erro + ((char)c);

			}
			error.close();
			//TODO retirar isso quando retirar a opção de chmod
			if (erro != "" && !cmd.contains("chmod"))
				System.err.println("Error Message in Prover: " + erro);
		}
		else {
			while ((c = in.read()) != -1) {
				output = output + ((char)c);
			}
			in.close();
		}
		return output;
	}

	/**
	 *  Metodo auxiliar 
	 * @param str
	 * @return
	 */
	private static String[] returnLines(String str)
	{
		StringTokenizer st = new StringTokenizer(str,lineSep);
		List<String> list = new Vector<String>();
		while (st.hasMoreTokens()) {
			list.add(st.nextToken());
		}
		String[] retorno = list.toArray(new String[0]);        
		return retorno;
	}

	// gerar arquivo SMT puro

	/**
	 * Pega um predicado, gera o arquivo smt do veriT, executa
	 * o veriT e pega a resposta.
	 * @param name nome do arquivo
	 */
	public static ResultStatusOP proveByVeriT (Pred pred){

		String name =  "generatedFile.smt";
		name = /*ZB2SMTUtils.getAbsPathforWindows()*/ abspath + generatedFilePath + name;

		// pega o path correspondente
		if (generateVeritFile(name, pred)){
			return executeMultiPlataformProver(name,veriTpathLinux,veriTpathWin);
		}
		else{
			/* Não ha como passar o predicado para a estrutua Bol*/
			return ResultStatusOP.UNDETERMINED; 
		}
	}



	public static ResultStatusOP proveByYices (Pred pred){

		String name =  "generatedFile.smt";
		name = /*ZB2SMTUtils.getAbsPathforWindows()*/ abspath + generatedFilePath + name;

		try {
			generatePureSMTFile(name, pred);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.err.println("Error in creating smt file");
			return ResultStatusOP.UNDETERMINED;
		}
		return executeMultiPlataformProver(name, yicesPathLinux, yicesPathWin);

	}


	public static ResultStatusOP proveByYices (LinkedList<POLabel> pos){

		String name =  "generatedFile.smt";
		name = (new File("")).getAbsolutePath() + generatedFilePath + name;	

		try {
			generatePureSMTFile(name, pos);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.err.println("Error in creating smt file");
			return ResultStatusOP.UNDETERMINED;
		}
		return executeMultiPlataformProver(name, yicesPathLinux, yicesPathWin);

	}



	public static ResultStatusOP proveByZ3(Pred pred){


		if (isWindows()){
			try {
				generatePureSMTFile("generatedFile.smt", pred);
				return executeZ3("generatedFile.smt");
			}
			catch (Exception e) {
				e.printStackTrace();
				System.err.println("Error in creating smt file");
				return ResultStatusOP.UNDETERMINED;
			}

		}
		else{
			System.err.println("Our current version of z3 is only available on Windows");
			return ResultStatusOP.UNDETERMINED;
		}

	}


	public static ResultStatusOP proveByCVC3(Pred pred){
		return proveByLinuxSolver(pred, cvc3path);
	}

	public static ResultStatusOP proveByAltErgo(Pred pred){
		return proveByLinuxSolver(pred, altErgopath);
	}


	private static ResultStatusOP proveByLinuxSolver(LinkedList<POLabel> pos, String pathSolver){

		String name =  "generatedFile.smt";
		name = abspath + generatedFilePath + name;

		if (isLinux()){
			try {
				generatePureSMTFile(name, pos);
				return executeLinuxProver(name, pathSolver);
			}
			catch (Exception e) {
				e.printStackTrace();
				System.err.println("Error in creating smt file");
				return ResultStatusOP.UNDETERMINED;
			}

		}
		else{
			System.err.println("Our current version of CVC3 is only available on Linux");
			return ResultStatusOP.UNDETERMINED;
		}

	}



	private static ResultStatusOP proveByLinuxSolver(Pred pred, String pathSolver){

		String name =  "generatedFile.smt";
		name = (new File("")).getAbsolutePath() + generatedFilePath + name;

		if (isLinux()){
			try {
				generatePureSMTFile(name, pred);
				return executeLinuxProver(name, pathSolver);
			}
			catch (Exception e) {
				e.printStackTrace();
				System.err.println("Error in creating smt file");
				return ResultStatusOP.UNDETERMINED;
			}

		}
		else{
			System.err.println("Our current version of CVC3 is only available on Linux");
			return ResultStatusOP.UNDETERMINED;
		}

	}

	public static ResultStatusOP proveByCVC3( LinkedList<POLabel> pos){
		return proveByLinuxSolver(pos, cvc3path);
	}

	public static ResultStatusOP proveByAltErgo( LinkedList<POLabel> pos){
		return proveByLinuxSolver(pos, altErgopath);
	}


	public static ResultStatusOP proveByZ3( LinkedList<POLabel> pos){

		if (isWindows()){
			try {
				generatePureSMTFile("generatedFile.smt", pos);
				return executeZ3("generatedFile.smt");
			}
			catch (Exception e) {
				e.printStackTrace();
				System.err.println("Error in creating smt file");
				return ResultStatusOP.UNDETERMINED;
			}

		}
		else{
			System.err.println("Our current version of z3 is only available on Windows");
			return ResultStatusOP.UNDETERMINED;
		}

	}


	private static String getAbsPathforWindows() {
		String result= "";
		String strPath  = abspath.substring(1,ZB2SMTUtils.abspath.length() - defaultStr.length() -1);

		for (int i=0 ; i< strPath.length();i++){

			char c = strPath.charAt(i);
			if (c == '/'){
				result += '\\';
			}
			else if (c == '%'){
				result += ' ';
				i = i +2;
			}
			else{
				result+= c;
			}
		}
		return result;
	}

	private static boolean isLinux() {
		String osName = System.getProperty("os.name");
		if (osName.contains("Linux")) {
			return true;
		}
		return false;
	}

	private static boolean isWindows() {
		String osName = System.getProperty("os.name");
		if (osName.startsWith("Windows")) {
			return true;
		}
		return false;
	}
	//TODO verificar esse metodo : Valerio
	/**
	 * Pega um predicado, gera o arquivo smt do harvey, executa
	 * o Harvey e pega a resposta.
	 */
	public static ResultStatusOP proveByVeriT ( LinkedList<POLabel> pos){

		String name =  "generatedFile.smt";
		name = abspath + generatedFilePath + name;

		generateVeritFile(name, pos);
		return executeMultiPlataformProver(name, veriTpathLinux, veriTpathWin);
	}

	/**
	 * Gera a string que representa o arquivo smt correspondente que contem o
	 * predicado pred
	 * @param pred Predicado
	 * @return
	 */
	public static StringBuffer generateVeritText(Pred pred){
		StringBuffer body = new StringBuffer();
		ConvertCircus2Bol convert = new ConvertCircus2Bol();

		Predicate pred2 = (Predicate)convert.visitTerm(pred);

		POLabel po = new POLabel("PO", pred2, null, convert.getIdentificadores(), null);
		LinkedList<POLabel> pos = new LinkedList<POLabel>();
		pos.add(po);

		POPrintVerit harveyPrinter = new POPrintVerit();
		harveyPrinter.toPrintInBuffer(pos, body, "     ", "file");
		return body;
	}

	/**
	 * Gera a String  correspondente do SMT puro
	 * @param pred
	 * @return
	 */
	public static StringBuffer generatePureSMTText(Pred pred){

		String pathFile =  "generatedFile.smt";
		pathFile = abspath + generatedFilePath + pathFile;
		// pega o path correspondente
		if (generateVeritFile(pathFile, pred)){

			StringBuffer result = new StringBuffer();

			String os = System.getProperty("os.name");
			String cmd;

			if ( os!= null && os.startsWith(WIN_ID)){
				pathFile = "\"" + pathFile + "\"";
				cmd = abspath + veriTpathWin + veriTparamPureSMTFileCmd+ pathFile;
			}
			else {
				/*
				 * Para que o harvey tenha permissao de ser executado
				 */

				try {
					executeCmd("chmod +x " + abspath + veriTpathLinux);
				} catch (IOException e) {
					e.printStackTrace();
				}

				cmd = abspath + veriTpathLinux  + veriTparamPureSMTFileCmd + pathFile;
			}


			try {
				String output = executeCmd(cmd);
				String[] lines = returnLines(output);

				for (int i=1; i< lines.length;i++){
					result.append(lines[i] + "\n");
				}
				return result;
			} catch (IOException e) {
				e.printStackTrace();
			}

		}

		return null;
	}


	/**
	 * Gera a String  correspondente do SMT puro
	 * @param pred
	 * @return
	 */
	public static StringBuffer generatePureSMTText(LinkedList<POLabel> predicados){

		String pathFile =  "generatedFile.smt";
		pathFile = abspath + generatedFilePath + pathFile;
		// pega o path correspondente
		generateVeritFile(pathFile, predicados);

		StringBuffer result = new StringBuffer();

		String os = System.getProperty("os.name");
		String cmd;

		if ( os!= null && os.startsWith(WIN_ID)){
			pathFile = "\"" + pathFile + "\"";
			cmd = abspath + veriTpathWin + veriTparamPureSMTFileCmd+ pathFile;
		}
		else {
			/*
			 * Para que o harvey tenha permissao de ser executado
			 */

			try {
				executeCmd("chmod +x " + abspath + veriTpathLinux);
			} catch (IOException e) {
				e.printStackTrace();
			}

			cmd = abspath + veriTpathLinux  + veriTparamPureSMTFileCmd + pathFile;
		}


		try {
			String output = executeCmd(cmd);
			String[] lines = returnLines(output);

			for (int i=1; i< lines.length;i++){
				result.append(lines[i] + "\n");
			}
			return result;
		} catch (IOException e) {
			e.printStackTrace();
		}


		return null;
	}


	/**
	 * Pega um predicado em Circus e gera o arquivo SMT Puro correspondente
	 * no arquivo path
	 * @param path
	 * @param pred
	 */
	public static void generatePureSMTFile(String path, Pred pred){

		String pathFile =  "generatedFile.smt";
		pathFile = abspath + generatedFilePath + pathFile;
		// pega o path correspondente
		if (generateVeritFile(pathFile, pred)){

			StringBuffer body = new StringBuffer();

			String os = System.getProperty("os.name");
			String cmd;

			if ( os!= null && os.startsWith(WIN_ID)){
				pathFile = "\"" + pathFile + "\"";
				cmd = abspath + veriTpathWin + veriTparamPureSMTFileCmd+ pathFile;
			}
			else {
				/*
				 * Para que o harvey tenha permissao de ser executado
				 */

				try {
					executeCmd("chmod +x " + abspath + veriTpathLinux);
				} catch (IOException e) {
					e.printStackTrace();
				}

				cmd = abspath + veriTpathLinux  + veriTparamPureSMTFileCmd + pathFile;
			}


			try {
				String stream = executeCmd(cmd);
				String[] lines = returnLines(stream);

				for (int i=1; i< lines.length;i++){
					body.append(lines[i] + "\n");
				}

				File output;
				if (path.endsWith(".smt")) {
					output = new File(path);
				}
				else{
					output = new File(path + ".smt");
				}
				/* Escrevendo o arquivo */
				FileOutputStream out = new FileOutputStream(output);
				byte [] data = body.toString().getBytes();
				out.write(data);
				out.close();

			} catch (IOException e) {
				e.printStackTrace();
			}

		}
	}

	/**
	 * Pega um predicado em Circus e gera o arquivo SMT Puro correspondente
	 * no arquivo path
	 * @param path
	 * @param pred
	 */
	public static void generatePureSMTFile(String path, LinkedList<POLabel> predicados){

		String pathFile =  "generatedFile.smt";
		pathFile = abspath + generatedFilePath + pathFile;
		// pega o path correspondente
		generateVeritFile(pathFile, predicados);

		StringBuffer body = new StringBuffer();

		String os = System.getProperty("os.name");
		String cmd;

		if ( os!= null && os.startsWith(WIN_ID)){
			pathFile = "\"" + pathFile + "\"";
			cmd = abspath + veriTpathWin + veriTparamPureSMTFileCmd+ pathFile;
		}
		else {
			/*
			 * Para que o harvey tenha permissao de ser executado
			 */
			try {
				executeCmd("chmod +x " + abspath + veriTpathLinux);
			} catch (IOException e) {
				e.printStackTrace();
			}

			cmd = abspath + veriTpathLinux  + veriTparamPureSMTFileCmd + pathFile;
		}


		try {
			String stream = executeCmd(cmd);
			String[] lines = returnLines(stream);

			for (int i=1; i< lines.length;i++){
				body.append(lines[i] + "\n");
			}

			File output;
			if (path.endsWith(".smt")) {
				output = new File(path);
			}
			else{
				output = new File(path + ".smt");
			}
			/* Escrevendo o arquivo */
			FileOutputStream out = new FileOutputStream(output);
			byte [] data = body.toString().getBytes();
			out.write(data);
			out.close();

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Tries to prove the predicate in Z by the solvers listed in the config file
	 * @param pred
	 * @return 
	 */
	public static ResultStatusOP executeAccordingConfig (Pred pred) {
		try {
			String text = generatePureSMTText(pred).toString();

			ProofClient client = new ProofClient(text);

			return getResultFromSolversAnswerString(client.proveRemotely());
		}
		catch (Exception e) {
			System.err.println("Maybe the Server was not initialized ");
			return ResultStatusOP.UNDETERMINED;
		}
	}


	/**
	 * Tries to prove the predicate in B by the solvers listed in the config file
	 * @param pred
	 * @return 
	 */
	public static ResultStatusOP executeAccordingConfig (LinkedList<POLabel> predicates) {

		try{
			String text = generatePureSMTText(predicates).toString();
			ProofClient client = new ProofClient(text);

			return getResultFromSolversAnswerString(client.proveRemotely());
		}
		catch (Exception e) {
			e.printStackTrace();
			return ResultStatusOP.UNDETERMINED;
		}
	}


}









